from typing import List, Tuple
import math

import pandas as pd
import matplotlib.pyplot as plt

from wmutils.pandas.figures.base import safe_save_fig


def create_figure(
    df: pd.DataFrame,
    x_key: str,
    synth_key: str,
    real_key: str,
    y_axis: str,
    output_path: "str | None" = None,
):
    # Create a figure and axis
    plt.figure()
    ax = plt.axes()

    # Plot the first line (sine function)
    ax.plot(df[x_key], df[synth_key], label="Theoretical", color="blue")

    # Plot the second line (cosine function)
    ax.plot(df[x_key], df[real_key], label="Experimental", color="red")
    upper = df[real_key] + df[f"std_{real_key}"]
    lower = df[real_key] - df[f"std_{real_key}"]
    ax.fill_between(df[x_key], upper, lower, facecolor="red", alpha=0.25)

    # Add labels and a legend
    ax.set_ylabel(y_axis)
    ax.set_xlabel(de_snake_case(x_key))
    # ax.set_title("Theoretical and experimental results for response latency")
    ax.legend()

    # Show the plot
    if output_path:
        print(f"{output_path=}")
        safe_save_fig(output_path)
    else:
        plt.show()


def create_split_multiline_figure(
    df: pd.DataFrame,
    split_key: str,
    x_key: str,
    y_key: str,
    y_key2: str,
    output_path: "str | None" = None,
):
    _, axes = plt.subplots(1, 2, figsize=(12, 4))
    exp_ax, synth_ax = axes[0], axes[1]

    splits = df[split_key].unique()

    for split in splits:
        sub_df = df[df[split_key] == split]
        label = f"{de_snake_case(split_key)} = {split}"
        exp_ax.plot(sub_df[x_key], sub_df[y_key], label=label)
        std_key = f"std_{y_key}"
        if std_key in sub_df.columns:
            upper = sub_df[y_key] + sub_df[std_key]
            lower = sub_df[y_key] - sub_df[std_key]
            exp_ax.fill_between(sub_df[x_key], lower, upper, alpha=0.25)

        synth_ax.plot(sub_df[x_key], sub_df[y_key2], label=label)
        std_key = f"std_{y_key2}"
        if std_key in sub_df.columns:
            upper = sub_df[y_key2] + sub_df[std_key]
            lower = sub_df[y_key2] - sub_df[std_key]
            synth_ax.fill_between(sub_df[x_key], lower, upper, alpha=0.25)

    exp_ax.legend()
    synth_ax.legend()

    plt.tight_layout()

    # Show the plot
    if output_path:
        print(f"{output_path=}")
        safe_save_fig(output_path)
    else:
        plt.show()


def create_multi_figure(
    df: pd.DataFrame,
    y_keys: List[str],
    x_key: str,
    split_key: str,
    y_label: str,
    output_path: "str | None" = None,
    num_cols: int = 2,
):
    if split_key:
        splits = df[split_key].unique()
    else:
        splits = [split_key]
    splits.sort()
    num_subplots = len(splits)

    num_rows = math.ceil(num_subplots / num_cols)
    width = 6 * num_cols
    height = 4 * num_rows
    _, axes = plt.subplots(num_rows, num_cols, figsize=(width, height))
    axes = axes.flatten()
    # axes = axes.flatten() if num_rows > 1 else [axes]
    for i, split in enumerate(splits):
        if split_key:
            tmp_df: pd.DataFrame = df[df[split_key] == split]
        else:
            tmp_df: pd.DataFrame = df
        ax: plt.Axes = axes[i]

        tmp_df = tmp_df.sort_values(x_key)
        for y_key in y_keys:
            ax.plot(tmp_df[x_key], tmp_df[y_key], label=de_snake_case(y_key))
            # tmp_df.plot(x_key, y_key, ax=ax)
            std_key = f"std_{y_key}"
            upper = tmp_df[y_key] + tmp_df[std_key]
            lower = tmp_df[y_key] - tmp_df[std_key]
            ax.fill_between(tmp_df[x_key], upper, lower, alpha=0.4)

        ax.set_xlabel(de_snake_case(x_key))
        ax.set_ylabel(y_label)
        ax.set_title(f"{de_snake_case(split_key)} = {split}")
        ax.legend()

    plt.tight_layout()

    # Show the plot
    if output_path:
        print(f"{output_path=}")
        safe_save_fig(output_path)
    else:
        plt.show()


def create_plot_comparisons(
    comparison_tuples: List[Tuple[str]],
    df: pd.DataFrame,
    x_column: str,
    output_path: "str | None" = None,
):
    # Calculate the number of subplots based on the length of comparison_tuples
    num_subplots = len(comparison_tuples)
    # Determine the number of rows and columns for the subplots
    num_rows = math.ceil(num_subplots / 2)  # Assuming 2 columns
    num_cols = 2  # Number of columns for the subplots

    # Create a larger figure with subplots
    width = 6 * num_cols
    height = 4 * num_rows
    _, axes = plt.subplots(num_rows, num_cols, figsize=(width, height))

    # Flatten the axes array if there is more than one row
    axes = axes.flatten() if num_rows > 1 else [axes]

    for i, columns in enumerate(comparison_tuples):
        # Select the current subplot
        ax: plt.Axes = axes[i]

        # Plot a line diagram for each pair of columns in the DataFrame
        for column in columns:
            ax.plot(df[x_column], df[column], label=de_snake_case(column))
            std_key = f"std_{column}"
            upper = df[column] + df[std_key]
            lower = df[column] - df[std_key]
            ax.fill_between(df[x_column], upper, lower, alpha=0.4)

        ax.set_xlabel(de_snake_case(x_column))
        ax.set_ylabel("CPU utilization")
        ax.legend()

    # Adjust layout to prevent overlapping titles
    plt.tight_layout()
    if output_path:
        print(f"{output_path=}")
        safe_save_fig(output_path)
    else:
        plt.show()


def de_snake_case(word: str) -> str:
    chunks = word.split("_")
    chunks = [chunk[0].upper() + chunk[1:] for chunk in chunks]
    word = " ".join(chunks)
    return word



from numbers import Number
from typing import Tuple

import pandas as pd


def normalize(x: Number, x_min: Number, x_max: Number) -> Number:
    x -= x_min
    x /= x_max - x_min
    return x


def normalize_field(df: pd.DataFrame, field: str) -> pd.DataFrame:
    min_x, max_x = min(df[field]), max(df[field])
    df.loc[:, field] = df[field].transform(lambda x: normalize(x, min_x, max_x))
    return df


def normalize_field_and_yield_min_max(
    df: pd.DataFrame, field: str
) -> Tuple[pd.DataFrame, Number, Number]:
    df = df.copy()
    min_x, max_x = min(df[field]), max(df[field])
    df.loc[:, field] = df[field].transform(lambda x: normalize(x, min_x, max_x))
    return df, min_x, max_x

def min_max_scale(_df: pd.DataFrame, scaled_fields: pd.Series):
    scaled_df = _df.copy()

    for feature in scaled_fields:
        feature_min = scaled_df[feature].min()
        feature_max = scaled_df[feature].max()
        feature_delta = feature_max - feature_min

        scaled_df[feature] = (
            scaled_df[feature].subtract(feature_min).divide(feature_delta)
        )

    return scaled_df


from datetime import datetime
import random


class Counter:
    """Simple tool for picking the next number in line."""

    def __init__(self, start_value: int = 42, increment: int = 1) -> None:
        self.__current_value = start_value
        self.__increment = increment

    def get_next(self):
        self.__current_value += self.__increment
        return self.__current_value


def lies_between(x, start, end) -> bool:
    return x >= start and x <= end


def lies_outside_timewindow(
    timestamp: str, start: datetime, end: datetime, time_format: str
) -> bool:
    """Returns true if the provided timestamp lies outside the provided window."""
    timestamp = datetime.strptime(timestamp, time_format)
    return timestamp < start and timestamp > end


def shuffled_range(start=0, end=100, step=1):
    lst = list(range(start, end, step))
    random.shuffle(lst)
    return lst



import unittest

from wmutils.collections import dict_access as d


class TestDictAccess(unittest.TestCase):

    def test_has_all_keys(self):
        my_dict = {"key_a": 123, "key_b": "asdf", "key_c": {789321}}

        self.assertTrue(d.has_all_keys(my_dict, ["key_a", "key_b", "key_c"]))
        self.assertTrue(d.has_all_keys(my_dict, ["key_a", "key_c"]))
        self.assertTrue(d.has_all_keys(my_dict, []))
        self.assertFalse(d.has_all_keys(my_dict, ["key_a", "key_b", "key_c", "key_d"]))
        self.assertFalse(d.has_all_keys(my_dict, ["key_d"]))

    def test_safe_add_list_element(self):
        my_dict = {}

        # Check 1: New element
        d.safe_add_list_element(my_dict, "key_a", "element_1")

        self.assertIn("key_a", my_dict)
        self.assertIsInstance(my_dict["key_a"], list)
        self.assertIs(len(my_dict["key_a"]), 1)

        # Check 2: pre-existing list
        my_dict = {"key_a": [0, 1, 2]}
        d.safe_add_list_element(my_dict, "key_a", 4)

        self.assertTrue(
            all(ele in my_dict["key_a"] for ele in range(3)),
            "Original element is missing.",
        )
        self.assertIn(4, my_dict["key_a"], "New element is missing")

    def test_safe_add_set_element(self):
        my_dict = {}

        # Check 1: New element
        d.safe_add_set_element(my_dict, "key_a", "element_1")

        self.assertIn("key_a", my_dict)
        self.assertIsInstance(my_dict["key_a"], set)
        self.assertIs(len(my_dict["key_a"]), 1)

        # Check 2: pre-existing set
        my_set = set()
        my_set.update(range(3))
        my_dict = {"key_a": my_set}
        d.safe_add_set_element(my_dict, "key_a", 4)

        self.assertTrue(
            all(ele in my_dict["key_a"] for ele in range(3)),
            "Original element is missing.",
        )
        self.assertIn(4, my_dict["key_a"], "New element is missing")

    def test_get_nested(self):
        my_dict = {"key_a": "value_a"}

        # Simple test
        self.assertEqual(d.get_nested(my_dict, ["key_a"]), "value_a")

        # Nested test.
        my_dict = {"key_a": {"key_b": {"key_c": 5}}}
        self.assertEqual(d.get_nested(my_dict, ["key_a", "key_b", "key_c"]), 5)

        # Simple missing key
        my_dict = {"key_a": 5}
        self.assertIsNone(d.get_nested(my_dict, ["key_b"]))

        # Nested missing key
        my_dict = {"key_a": {"key_b": {"key_c": 5}}}
        self.assertIsNone(d.get_nested(my_dict, ["key_a", "key_b", "key_d"]))

        # Test with list
        my_dict = {"key_a": {"key_b": [0, 1, 2, 3]}}
        res = d.get_nested(my_dict, ["key_a", "key_b"])
        self.assertIsNotNone(res)
        self.assertEqual(len(res), 4)
        self.assertTrue(all([ele in res for ele in range(4)]))

    def test_get_nested_many_simple_retrieval(self):
        # Simple test
        my_dict = {"key_a": "value"}
        res = d.better_get_nested_many(my_dict, ["key_a"])
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], "value")

        # Nested array at root test
        my_dict = [{"key_a": "value_0"}, {"key_a": "value_1"}, {"key_a": "value_2"}]
        res = d.better_get_nested_many(my_dict, ["key_a"])
        self.assertEqual(len(res), 3)
        self.assertTrue(all(f"value_{ele}" in res for ele in range(3)))

        # Nested array at sublevel
        my_dict = {
            "key_a": [{"key_b": "value_0"}, {"key_b": "value_1"}, {"key_b": "value_2"}]
        }
        res = d.better_get_nested_many(my_dict, ["key_a", "key_b"])
        self.assertEqual(len(res), 3)
        self.assertTrue(all(f"value_{ele}" in res for ele in range(3)))

    def test_get_nested_many_missing_key(self):
        # Simple missing key test no raise.
        my_dict = {"key_a": "value"}
        res = d.better_get_nested_many(my_dict, ["key_b"], False)
        self.assertEqual(len(res), 0)

        # Simple missing key raise
        my_dict = {"key_a": "value"}
        self.assertRaises(KeyError, d.better_get_nested_many, my_dict, ["key_b"], True)

        # Nested missing key
        my_dict = [{"key_a": "value_0"}, {"key_a": "value_1"}, {"key_b": "value_2"}]
        self.assertRaises(KeyError, d.better_get_nested_many, my_dict, ["key_a"], True)

        # Nested missing key at sublevel
        my_dict = {
            "key_a": [{"key_b": "value_0"}, {"key_b": "value_1"}, {"key_b": "value_2"}]
        }
        self.assertRaises(
            KeyError, d.better_get_nested_many, my_dict, ["key_a", "key_c"], True
        )

    def test_get_nested_many_complex_retrieval(self):
        # # Test tree.
        # my_dict = {
        #     "key_a": [
        #         {"key_b": [{"key_c": 1}, {"key_c": 2}, {"key_c": 3}]},
        #         {"key_b": [{"key_c": 4}, {"key_c": 5}, {"key_c": 6}]},
        #         {"key_b": [{"key_c": 7}, {"key_c": 8}, {"key_c": 0}]},
        #     ]
        # }

        # res = d.better_get_nested_many(my_dict, ["key_a", "key_b", "key_c"])
        # self.assertEqual(len(res), 9)
        # print(res)
        # self.assertTrue(all([ele in res for ele in range(10)]))
        pass



from typing import Iterator

import regex as re


def get_matching(collection: Iterator[str], pattern: str) -> Iterator[str]:
    """Returns all entries that match the expression."""

    def __is_pattern_match(element: str) -> bool:
        return re.match(pattern, element)

    return filter(__is_pattern_match, collection)


def get_non_matching(collection: Iterator[str], pattern: str) -> Iterator[str]:
    """Returns all entries that do NOT match the expression."""

    def __is_not_pattern_match(element: str) -> bool:
        return not re.match(pattern, element)

    return filter(__is_not_pattern_match, collection)



import pandas as pd


def reset_column_name_indices(df: pd.DataFrame):
    """Resets the suffix indices of dataframe columns;
    s.t. `name.1` becomes `name` if there is only one."""

    def __get_real_name(col: str) -> str:
        return col.split(".")[0]

    columns = df.columns
    unique_cols = set([__get_real_name(col) for col in columns])
    appearances = {col: 0 for col in unique_cols}
    new_cols = []
    for col in columns:
        real_name = __get_real_name(col)
        if appearances[real_name] == 0:
            new_cols.append(real_name)
        else:
            new_name = f"{real_name}.{appearances[real_name]}"
            new_cols.append(new_name)
        appearances[real_name] += 1

    was_updated = any(old != new for old, new in zip(df.columns, new_cols))
    if was_updated:
        print(f"Updated df columns to: {new_cols}")

    df.columns = new_cols



import io
import os

from typing import List, Iterator


class OpenMany:
    def __init__(self, file_paths: List[str], *args, **kwargs) -> None:
        self.file_paths: List[str] = file_paths
        self.files: List[io.IOBase] = [None] * len(file_paths)
        self.__args = args
        self.__kwargs = kwargs

    def __enter__(self) -> List[io.IOBase]:
        for index, file_path in enumerate(self.file_paths):
            self.files[index] = open(file_path, *self.__args, **self.__kwargs)
        return self.files

    def __exit__(self, type, value, traceback) -> None:
        for file in self.files:
            file.close()


def safe_makedirs(dirname: str):
    if not os.path.exists(dirname):
        os.makedirs(dirname)


def get_subfolders(parent_dir) -> Iterator[str]:
    """Returns the folders in a directory."""
    return [
        f"{parent_dir}/{name}"
        for name in os.listdir(parent_dir)
        if os.path.isdir(f"{parent_dir}/{name}")
    ]


def iterate_through_nested_folders(base_folder: str, max_depth: int) -> Iterator[str]:
    """Iterates through all folders with a certain depth from the specified base folder."""
    yield base_folder
    sub_folders = get_subfolders(base_folder)
    if max_depth > 0:
        for subfolder in sub_folders:
            for folder in iterate_through_nested_folders(subfolder, max_depth - 1):
                yield folder
    else:
        for folder in get_subfolders(base_folder):
            yield folder


def iterate_through_files_in_nested_folders(
    base_folder: str, max_depth: int
) -> Iterator[str]:
    """Iterates through all of the files that are in the folders with a
    certain depth from the specified base folder."""
    for folder in iterate_through_nested_folders(base_folder, max_depth):
        for file in os.listdir(folder):
            yield f"{folder}/{file}"



from dataclasses import dataclass
import unittest
import multiprocessing

from wmutils.multithreading import parallelize_tasks


THREAD_COUNT = 8
WORK_LOAD = 10000


@dataclass
class MyObject:
    task: int
    task_id: int
    worker_id: int
    total_tasks: int | str


def return_task(task, task_id, worker_id, total_tasks):
    return task


def return_dict(task, task_id, worker_id, total_tasks):
    return {
        "task": task,
        "task_id": task_id,
        "worker_id": worker_id,
        "total_tasks": total_tasks,
        "constant": "my_constant",
    }


def return_object(task, task_id, worker_id, total_tasks):
    return MyObject(task, task_id, worker_id, total_tasks)


class TestMultithreading(unittest.TestCase):
    def test_paralellize_tasks_termination(self):
        tasks = range(WORK_LOAD)

        def bar():
            results = parallelize_tasks(
                tasks, return_object, thread_count=THREAD_COUNT, return_results=True
            )

        p = multiprocessing.Process(target=bar)
        p.start()
        p.join(timeout=2)
        self.assertFalse(p.is_alive(), "Process did not terminate.")
        p.kill()

    def test_paralellize_tasks_no_results(self):
        tasks = range(WORK_LOAD)

        results = parallelize_tasks(
            tasks, return_task, thread_count=THREAD_COUNT, return_results=False
        )

        self.assertIsNone(results)

    def test_paralellize_tasks_with_primitive_results(self):
        tasks = range(WORK_LOAD)

        results = parallelize_tasks(
            tasks, return_task, thread_count=THREAD_COUNT, return_results=True
        )

        results = list(results)

        self.assertEqual(len(results), WORK_LOAD)

        tasks = set(tasks)
        results = set(results)
        intersect = tasks.intersection(results)
        self.assertEqual(len(intersect), WORK_LOAD, "Received incorrect results.")

    def test_paralellize_countable_tasks_with_dict_results(self):
        tasks = list(range(WORK_LOAD))

        results = parallelize_tasks(
            tasks, return_dict, thread_count=THREAD_COUNT, return_results=True
        )

        # Result size test
        results = list(results)
        self.assertEqual(len(results), WORK_LOAD)

        # Content test.
        valid_worker_ids = set(range(0, THREAD_COUNT))
        results = sorted(results, key=lambda element: element["task"])
        for result, task in zip(results, tasks):
            self.assertEqual(result["task"], task)
            self.assertIn(result["worker_id"], valid_worker_ids)
            self.assertEqual(result["total_tasks"], len(tasks))
            self.assertEqual(result["constant"], "my_constant")

    def test_paralellize_uncountable_tasks_with_dict_results(self):
        tasks = range(WORK_LOAD)

        results = parallelize_tasks(
            tasks, return_dict, thread_count=THREAD_COUNT, return_results=True
        )

        # Result size test
        results = list(results)
        self.assertEqual(len(results), WORK_LOAD)

        # Content test.
        valid_worker_ids = set(range(0, THREAD_COUNT))
        results = sorted(results, key=lambda element: element["task"])
        for result, task in zip(results, tasks):
            self.assertEqual(result["task"], task)
            self.assertIn(result["worker_id"], valid_worker_ids)
            self.assertEqual(result["total_tasks"], "unknown")
            self.assertEqual(result["constant"], "my_constant")

    def test_paralellize_tasks_with_object_results(self):
        tasks = range(WORK_LOAD)

        results = parallelize_tasks(
            tasks, return_object, thread_count=THREAD_COUNT, return_results=True
        )

        # Result size test
        results = list(results)
        self.assertEqual(len(results), WORK_LOAD)

        # Content test.
        valid_worker_ids = set(range(0, THREAD_COUNT))
        valid_task_ids = set(range(0, WORK_LOAD))
        used_task_ids = set()
        results = sorted(results, key=lambda element: element.task)
        for result, task in zip(results, tasks):
            self.assertIsInstance(result, MyObject)
            self.assertEqual(result.task, task)
            self.assertIn(result.worker_id, valid_worker_ids)
            self.assertEqual(result.total_tasks, "unknown")
            self.assertIn(result.task_id, valid_task_ids)
            self.assertNotIn(result.task_id, used_task_ids)

            used_task_ids.add(result.task_id)



import math
from typing import Iterator, List, Callable, TypeVar, Tuple, Dict
from numbers import Number

import numpy

T = TypeVar("T")


def ordered_chain(
    iterables: List[Iterator[T]], key: Callable[[T, T], Number]
) -> Iterator[Tuple[int, T]]:
    """
    Iterates through multiple generators in a chained fashion,
    iterating through them in an ordered fashion. Assumes the
    individual generators are sorted already.

    :param list[Generator[T]] iterables: The lists that are being chained.
    :param Callable[[T], Number] key: Method that is used for ordering
    iterable elements.
    """

    current_elements = [next(iterables[idx]) for idx in range(len(iterables))]
    stop_iterations = 0

    def __key_wrapper(entry):
        return math.inf if entry is None else key(entry)

    while stop_iterations != len(iterables):
        current_idx = numpy.argmin([__key_wrapper(ele) for ele in current_elements])
        yield current_idx, current_elements[current_idx]
        try:
            current_elements[current_idx] = next(iterables[current_idx])
        except StopIteration:
            stop_iterations += 1
            current_elements[current_idx] = None


def tuple_chain(
    iterator: Iterator[T], yield_first: bool = False, yield_last: bool = False
) -> "Iterator[Tuple[T | None, T | None]]":
    """Returns tuples of entries. Given [a, b, c, d], it outputs [(a,b), (b,c), (c,d)]"""
    if not isinstance(iterator, Iterator):
        iterator = iter(iterator)

    previous = None
    current = next(iterator)

    if yield_first:
        yield previous, current

    for entry in iterator:
        previous = current
        current = entry
        yield previous, current

    if yield_last:
        yield current, None


def chain_with_intermediary_callback(
    generator: Iterator[T], callback: Callable[[T], None]
) -> Iterator[T]:
    """Calls the specified function before yielding the entry like normal."""
    for entry in generator:
        callback(entry)
        yield entry


def stepped_enumerate(
    collection: Iterator[T], start: Number = 0, step: Number = 1
) -> Iterator[Tuple[Number, T]]:
    current = start
    for entry in collection:
        yield (current, entry)
        current += step


def merge_iterate_through_lists(
    collections: List[List[T]], sorting_key: Callable[[T], Number]
) -> Tuple[Number, Iterator[Dict[Number, T]]]:
    """
    Applies the same method used in MergeSort to iterate through various lists.
    If multiple entries have the same key, they are ALL yielded.
    Assumes that an individual collection has no duplicate sorting keys.
    """
    element_pointers = [0] * len(collections)
    counter = 0
    counter_max = sum([len(coll) for coll in collections])
    while counter < counter_max:
        # Finds the current elements with the lowest sorting key
        lowest = math.inf
        collection = {}
        for collection_index, elements in enumerate(collections):
            pointer = element_pointers[collection_index]
            if pointer >= len(elements):
                continue
            current: T = elements[pointer]
            element_value = sorting_key(current)
            # Creates new collection if a lower value is found.
            if element_value < lowest:
                lowest = element_value
                collection = {collection_index: current}
            # Appends collection if value is the same.
            elif element_value == lowest:
                collection[collection_index] = current
        # Updates guard and pointers.
        counter += len(collection)
        for collection_index in collection.keys():
            element_pointers[collection_index] += 1
        yield lowest, collection


def limit(iterator: Iterator[T], max_iterations: int) -> Iterator[T]:
    while max_iterations > 0:
        yield next(iterator)
        max_iterations -= 1


class RepeatingIterator:
    """Iterator that provides the option to repeat an entry."""
    def __init__(self, iterable):
        self.iterable = iter(iterable)
        self.last_item = None
        self.repeat_last = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.repeat_last:
            self.repeat_last = False
            return self.last_item
        else:
            self.last_item = next(self.iterable)
            return self.last_item

    def set_repeat_last(self, value=True):
        self.repeat_last = value



"""
Implements utility functions for interacting with dictionaries.
"""

from typing import Dict, Any, List, TypeVar, Set
from numbers import Number


K = TypeVar("K")
V = TypeVar("V")


def has_all_keys(collection: Dict[K, V], keys: List[K]) -> bool:
    """Returns true if all of the keys are present in the dictionary."""
    return all((key in collection for key in keys))


def safe_add_list_element(collection: Dict[K, List[V]], key: K, value: V):
    """
    Adds the new element to the list of elements in corresponding
    to the key if a list exists already, otherwise it creates a new list.
    :param dictionary: The dictionary to which the element is added.
    :param key: The key of the new element.
    :param value: The new element.
    """
    if key not in collection:
        collection[key] = list()
    collection[key].append(value)


def safe_add_set_element(collection: Dict[K, Set[V]], key: K, value: V):
    """Adds the new element to the set of elements in corresponding
    to the key if a set exists already, otherwise it creates a new set.
    :param collection: The dictionary to which the element is added.
    :param key: The key of the new element.
    :param value: The new element.
    """
    if key not in collection:
        collection[key] = set()
    collection[key].add(value)


def get_nested(collection: Dict[K, V], key: List[K]) -> "V | None":
    """
    Returns value corresponding to the key by recursively
    searching in the given dictionary. It returns `None`
    when the key is invalid.

    :params obj: The used dictionary.
    :params key: The query key.
    :return: The value, if such a value exists.
    """

    current = collection
    for key_element in key:
        if not key_element in current:
            return None
        current = current[key_element]
    return current


def get_nested_many(obj: dict, key: List[str]) -> "List[Any] | Any | None":
    """Same idea as ``get_nested``, however, when a variable is a list it iterates through all of them."""
    current = obj
    for key_index, key_element in enumerate(key):
        if isinstance(current, list):
            return [get_nested_many(element, key[key_index:]) for element in current]
        if not key_element in current:
            return None
        current = current[key_element]
    return current


# BUG: Query with Dict[Dict[List[List]]], which should yield a list of lists, doesn't work.
def better_get_nested_many(
    collection: Dict[Any, Any], nested_key: List[Any], raise_on_missing_key: bool = True
) -> List[Any]:
    """
    Same thing as ``get_nested_many`` but always returns a list.
    """

    current = collection
    for key_index, key_element in enumerate(nested_key):
        if isinstance(current, List):
            all_inner = [
                better_get_nested_many(
                    element, nested_key[key_index:], raise_on_missing_key
                )
                for element in current
            ]
            combined = []
            for inner in all_inner:
                combined.extend(inner)
            return combined
        elif not key_element in current:
            if raise_on_missing_key:
                raise KeyError(f"Missing key {key_element} in object {current}.")
            return []
        else:
            current = current[key_element]
    if isinstance(current, List):
        return current
    else:
        return [current]


def safe_get(collection: Dict[K, V], key: K, default: "V | None" = None) -> V:
    """Returns the value of the queried key, or the default if the key is not present."""
    if key in collection:
        return collection[key]
    return default


def subtract_dict(
    collection_a: Dict[K, Number], collection_b: Dict[K, Number]
) -> Dict[K, Number]:
    """
    Subtracts the values of one dict from another: `set_a - set_b`.
    :param collection_a: The dictionary subtracted from.
    :param collection_b: The dictionary that's subtracted.
    """
    key_intersect = set(collection_a.keys()).intersection(collection_b.keys())
    if len(key_intersect) != len(collection_a) or len(key_intersect) != len(
        collection_b
    ):
        raise ValueError("Elements don't have the same keys.")
    return {key: collection_a[key] - collection_b[key] for key in key_intersect}


def add_dict(
    collection_a: Dict[K, Number], collection_b: Dict[K, Number]
) -> Dict[K, Number]:
    """
    Subtracts the values of one dict from another: `set_a - set_b`.
    :param collection_a: The dictionary subtracted from.
    :param collection_b: The dictionary that's subtracted.
    """
    key_intersect = set(collection_a.keys()).intersection(collection_b.keys())
    if len(key_intersect) != len(collection_a) or len(key_intersect) != len(
        collection_b
    ):
        raise ValueError("Elements don't have the same keys.")
    return {key: collection_a[key] + collection_b[key] for key in key_intersect}


def invert_dict(collection: Dict[K, V]) -> Dict[V, K]:
    """
    Inverts the dictionary such that the keys become
    its values and vice versa. Assumes all values are unique.
    """
    return {value: key for key, value in collection.items()}



from typing import Iterator, Any, Callable, List, Sequence


def resolve_callables_in_list(
    coll: "Iterator[Any | Callable]", *args, **kwargs
) -> Iterator[Any]:
    for entry in coll:
        if isinstance(entry, Callable):
            entry = entry(*args, **kwargs)
        yield entry


def safe_index(list: List, entry: object) -> int:
    try:
        return list.index(entry)
    except ValueError:
        return -1


def flatten(iterator: "Iterator[Iterator | Any]") -> Iterator[Any]:
    """
    Flattens Iterator with nested Iterators.
    """

    for element in iterator:
        if (
            isinstance(element, Iterator) or isinstance(element, Sequence)
        ) and not isinstance(element, str):
            for inner_element in flatten(element):
                yield inner_element
        else:
            yield element



from setuptools import setup, find_packages

version = "0.13.0"

setup(
    name="wmutils",
    version=version,
    author="Willem Meijer",
    author_email="me@wmeijer.com",
    description="Contains a diverse set of utility methods I use in Python.",
    packages=find_packages(),
    url="https://github.com/wmeijer221/my_python_utils",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import typing

import matplotlib.pyplot as plt
import pandas as pd

from wmutils.pandas.figures.base import safe_save_fig


DEFAULT_PLOT_STYLE: str = "ggplot"
DEFAULT_OUTPUT_FORMAT: "typing.Iterator[str] | str" = ["png", "pdf"]


def __save_plot(
    figure_output_path: str,
    display_output: bool,
    output_format: "typing.Iterator[str] | str",
):
    if not figure_output_path is None:
        print(f"{figure_output_path=}")
        safe_save_fig(figure_output_path, display_output, output_format)
    elif display_output:
        plt.show()


def make_line_chart(
    df: pd.DataFrame,
    x_key: str,
    y_keys: typing.List[str],
    x_label: str,
    y_label: str,
    line_labels: typing.List[str],
    y_std_keys: "typing.List[str | None] | None" = None,
    figure_output_path: "str | None" = None,
    legend_position: "typing.Tuple[str, typing.Tuple[float, float]] | None" = None,
    output_type: "str | typing.Iterator[str]" = DEFAULT_OUTPUT_FORMAT,
    display_output: bool = False,
    style: str = DEFAULT_PLOT_STYLE,
):
    """Creates a line plot with `ggplot` formatting."""

    with plt.style.context(style):
        # Create a figure and axis
        plt.figure()
        ax = plt.axes()

        # Plots lines.
        for index, y_key in enumerate(y_keys):
            # Selects label.
            label = line_labels[index]

            # Plots line
            ax.plot(df[x_key], df[y_key], label=label)

            # Plots std.
            if not y_std_keys is None:
                y_std_key = y_std_keys[index]
                if y_std_key in df.columns:
                    std = df[y_std_key]
                else:
                    std = 0
                upper = df[y_key].add(std)
                lower = df[y_key].subtract(std)
                ax.fill_between(df[x_key], upper, lower, alpha=0.25)

        ax.set_ylabel(y_label)
        ax.set_xlabel(x_label)

        # Sets legend
        if legend_position is None:
            ax.legend()
        else:
            ax.legend(loc=legend_position[0], bbox_to_anchor=legend_position[1])

        __save_plot(figure_output_path, display_output, output_type)


def make_split_line_chart(
    df: pd.DataFrame,
    x_key: str,
    split_key: str,
    y_key: str,
    x_label: str,
    y_label: str,
    build_split_label: callable,
    y_std_key: "str | None" = None,
    figure_output_path: "str | None" = None,
    legend_position: "typing.Tuple[str, typing.Tuple[float, float]] | None" = None,
    display_output: bool = False,
    output_format: "str | typing.Iterator[str]" = DEFAULT_OUTPUT_FORMAT,
    style: str = DEFAULT_PLOT_STYLE,
):
    """Creates a multiline chart using different dataframe
    splits in the `ggplot` style."""
    with plt.style.context(style):
        plt.figure()
        ax = plt.axes()

        # Creates  splits.
        splits = df[split_key].unique()
        for split in splits:
            sub_df = df[df[split_key] == split]

            # Plots line.
            label = build_split_label(split)
            ax.plot(sub_df[x_key], sub_df[y_key], label=label)

            # Plots std.
            if not y_std_key is None:
                if y_std_key in sub_df.columns:
                    std = sub_df[y_std_key]
                else:
                    std = 0
                upper = sub_df[y_key].add(std)
                lower = sub_df[y_key].subtract(std)
                ax.fill_between(sub_df[x_key], upper, lower, alpha=0.25)

        ax.set_ylabel(y_label)
        ax.set_xlabel(x_label)

        # Sets legend
        if legend_position is None:
            ax.legend()
        else:
            ax.legend(loc=legend_position[0], bbox_to_anchor=legend_position[1])

        __save_plot(figure_output_path, display_output, output_format)



import os
from typing import Iterator

import matplotlib.pyplot as plt


def safe_save_fig(
    output_path, show_figure: bool = True, output_format: "str | Iterator[str]" = "png"
):
    """Helper method to safe figures in a potentially non-existent directory."""
    dir_name = os.path.dirname(output_path)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    # Strips ext from path.
    base_output_path = ".".join(output_path.split(".")[:-1])

    if isinstance(output_format, str):
        output_format = [output_format]
    for element in output_format:
        real_output_path = f"{base_output_path}.{element}"
        plt.savefig(real_output_path, dpi=400, format=element)

    if not show_figure:
        plt.close()



"""
Contains utility scripts for multithreading related tasks.
"""

import logging
import multiprocessing
from typing import Callable, Iterator, TypeVar, Any, List


T = TypeVar("T")
R = TypeVar("R")


class SimpleConsumer(multiprocessing.Process):
    """Simple consumer thread."""

    class TerminateTask:
        """When received by the simple consumer, it terminates."""

    def __init__(
        self,
        on_message_received: Callable,
        task_list: multiprocessing.JoinableQueue,
        worker_index: int,
        result_queue: multiprocessing.Queue,
        consumer_name: str = "SimpleConsumer",
        *args,
        **kwargs,
    ) -> None:
        super().__init__()

        self._on_message_received = on_message_received
        self._task_list = task_list
        self._worker_index = worker_index
        self._result_queue = result_queue
        self._args = args
        self._kwargs = kwargs
        self._consumer_name = f"{consumer_name}-{worker_index}"

    def run(self) -> None:
        """Implements simple execution lifecycle."""
        for task in self._my_tasks():
            self._execute_task(task)

    def _my_tasks(self) -> Iterator[R]:
        """Yields new tasks as long as no `TerminateTask` is found."""
        has_terminated = False
        while not has_terminated:
            task = self._task_list.get()
            if not isinstance(task, SimpleConsumer.TerminateTask):
                yield task
            else:
                has_terminated = True

    def _execute_task(self, task: R):
        """Attempts to execute the task."""
        try:
            task_kwargs = {
                **self._kwargs,
                **task,
                "worker_id": self._worker_index,
            }
            result = self._on_message_received(*self._args, **task_kwargs)
            if not self._result_queue is None:
                self._result_queue.put(result)
        except Exception as ex:
            logging.warning(f"{self._consumer_name}: Failed with entry {task}: {ex}.")
            raise


class ExecutorService:
    def __init__(
        self,
        thread_count: int = 1,
        return_results: bool = False,
        use_early_return_results: bool = True,
        *args,
        **kwargs,
    ):
        """
        Constructs a simple executor service.
        :param thread_count: The number of used threads (min. 1).
        :param return_results: Whether the return value of the
        submitted tasks should be stored.
        :param use_early_return_results: If there are return values,
        whether these should be collected before threads are joined.
        :param *args, **kwargs: Any other parameters that are passed
        to the worker threads.
        """

        if thread_count < 1:
            raise ValueError("You can't have less than one thread.")

        self._thread_count = thread_count
        self._return_results = return_results
        self._use_early_return_results = use_early_return_results
        self._args = args
        self._kwargs = kwargs

        self._worklist = multiprocessing.JoinableQueue()
        self._workers: list[SimpleConsumer] = [None] * thread_count
        self._result_queue = multiprocessing.Queue() if return_results else None
        self._early_return_results: List[R] | None = None

    def do_task(self, task_callable, targs, tkwargs, *args, **kwargs):
        return task_callable(*targs, *args, **tkwargs, **kwargs)

    def start(self):
        """Initializes worker threads."""
        for index in range(self._thread_count):
            worker = SimpleConsumer(
                self.do_task,
                self._worklist,
                index,
                result_queue=self._result_queue,
                *self._args,
                **self._kwargs,
            )
            worker.start()
            self._workers[index] = worker

    def submit(self, task_callable: Callable[[T, Any], R], *targs, **tkwargs):
        """
        Submits a task.
        :param task_callable: the method that is executed by the thread.
        :param *targs, **tkwargs: Any other parameter that is passed to the method.
        """
        task_wrapper = {
            "task_callable": task_callable,
            "targs": targs,
            "tkwargs": tkwargs,
        }
        self._worklist.put(task_wrapper)

    def stop(self):
        """
        Terminates all the workers, collects their results
        early (if specified in the constructor), and joins
        them. This method is blocking.
        """
        # Kills workers.
        for _ in range(self._thread_count):
            self._worklist.put(SimpleConsumer.TerminateTask())

        # Collects results early if desired.
        if self._return_results and self._use_early_return_results:
            self._early_return_results = list(self.get_results_iter())

        # Waits until workers terminate.
        for worker in self._workers:
            worker.join()

    def get_results_iter(self) -> Iterator[R] | None:
        """
        Yields a result iterator, if there is one.
        """
        if self._early_return_results:
            raise ValueError(
                "Constructor parameter `early_return_results` can't be true when using this method."
            )
        if not self._return_results:
            return
        while not self._result_queue.empty() or not self._worklist.empty():
            yield self._result_queue.get()

    def get_results(self) -> List[R] | None:
        """
        Yields a list of results if there are any.
        """
        if not self._return_results:
            return None
        if self._use_early_return_results and not self._early_return_results is None:
            return self._early_return_results
        return list(self.get_results_iter())


def parallelize_tasks(
    tasks: Iterator[T],
    on_task_received: Callable[[T, int, int, int | str], R],
    thread_count: int = 1,
    return_results: bool = False,
    use_early_return_results: bool = True,
    *args,
    **kwargs,
) -> Iterator[R] | List[R] | None:
    """
    Implements basic scheme for emberassingly parallel
    work; i.e., data parallelism.
    :param tasks: The tasks that are executed.
    :param on_message_received: Callable that processes a task.
    It receives named parameters: `task`, `task_id`, `worker_id`, `total_tasks`
    and any other parameter that is passed in `*args` or `**kwargs`.
    :param thread_count: The number of used threads.
    :param return_results: Whether the results of `on_message_received`
    should be kept.
    :param use_early_return_results: If there are return values,
    whether these should be collected before threads are joined.
    :return: If `return_results` is set to True, it returns the results,
    otherwise, it returns `None`.
    """

    executor = ExecutorService(
        thread_count, return_results, use_early_return_results, *args, **kwargs
    )
    executor.start()
    total_tasks = len(tasks) if isinstance(tasks, list) else "unknown"
    for task_id, task in enumerate(tasks):
        executor.submit(
            task_callable=on_task_received,
            task=task,
            task_id=task_id,
            total_tasks=total_tasks,
        )
    executor.stop()
    if use_early_return_results:
        return executor.get_results()
    else:
        return executor.get_results_iter()
